///////////////////////////////////////////////////////////////////////////////
//
// File: MappingExtrapolate.h
//
// For more information, please see: http://www.nektar.info
//
// The MIT License
//
// Copyright (c) 2006 Division of Applied Mathematics, Brown University (USA),
// Department of Aeronautics, Imperial College London (UK), and Scientific
// Computing and Imaging Institute, University of Utah (USA).
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// Description: Abstract base class for MappingExtrapolate.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef NEKTAR_SOLVERS_MAPPINGEXTRAPOLATE_H
#define NEKTAR_SOLVERS_MAPPINGEXTRAPOLATE_H

#include <GlobalMapping/Mapping.h>
#include <IncNavierStokesSolver/EquationSystems/StandardExtrapolate.h>

namespace Nektar
{
//--------
// Mapping Extrapolate
// --------

class MappingExtrapolate;

typedef std::shared_ptr<MappingExtrapolate> MappingExtrapolateSharedPtr;

class MappingExtrapolate : public StandardExtrapolate
{
public:
    friend class MemoryManager<MappingExtrapolate>;

    /// Creates an instance of this class
    static ExtrapolateSharedPtr create(
        const LibUtilities::SessionReaderSharedPtr &pSession,
        Array<OneD, MultiRegions::ExpListSharedPtr> &pFields,
        MultiRegions::ExpListSharedPtr &pPressure, const Array<OneD, int> &pVel,
        const SolverUtils::AdvectionSharedPtr &advObject)
    {
        ExtrapolateSharedPtr p =
            MemoryManager<MappingExtrapolate>::AllocateSharedPtr(
                pSession, pFields, pPressure, pVel, advObject);
        return p;
    }

    /// Name of class
    static std::string className;

    void v_CorrectPressureBCs(const Array<OneD, NekDouble> &pressure) override;

    void v_CalcNeumannPressureBCs(
        const Array<OneD, const Array<OneD, NekDouble>> &fields,
        const Array<OneD, const Array<OneD, NekDouble>> &N,
        NekDouble kinvis) override;

    ~MappingExtrapolate() override = default;

protected:
    MappingExtrapolate(const LibUtilities::SessionReaderSharedPtr pSession,
                       Array<OneD, MultiRegions::ExpListSharedPtr> pFields,
                       MultiRegions::ExpListSharedPtr pPressure,
                       const Array<OneD, int> pVel,
                       const SolverUtils::AdvectionSharedPtr advObject);

    // Mapping object
    GlobalMapping::MappingSharedPtr m_mapping;

    Array<OneD, NekDouble> m_bcCorrection;

    // Flags defining if pressure and viscous mapping terms
    // should be treated implicitly
    bool m_implicitPressure;
    bool m_implicitViscous;
    // Relaxation parameters for pressure
    //       system (when solved iteratively)
    NekDouble m_pressureRelaxation;
};
} // namespace Nektar

#endif
